@@ -254,6 +254,48 @@ static void debug_attr(META_SPC *s, META_ITEM *i, char *blk, char *p,
 }
 
 
+/* Helper for meta_decode; creates new A/V pair using attribute i according
+ * to its type, from raw data at val, vallen. */
+
+static META_AV *addavfromraw(META_AV **head, META_AV **tail,
+			 META_ITEM *i, char *p, META_ORD l, int copystrings)
+{
+    META_AV *av;
+
+    av = (META_AV *)malloc(sizeof(META_AV)); if (!av) goto error;
+    memset(av, 0, sizeof(META_AV));
+
+    av->i = i;
+
+    if (MT_ISORD(i->val_type)) {
+	    av->l = getord(p, l);
+    } else if (!copystrings) {
+	    av->l = l; 
+	    av->p = p;
+    } else {
+	    /* Create a copy for all strings on the heap
+	     * instead of having them point in the source 
+	     * buffer. Used by langvm when DICTDEC uses
+	     * a temporary string as the source. */
+	    av->l = l; 
+	    av->p = (char *)malloc(l);
+	    if (!av->p) { free(av); goto error; }
+	    memcpy(av->p, p, av->l); 
+	    av->flags = AV_FREE_P; 
+    }
+
+    if (!*head) *head = av;
+    else (*tail)->next = av;
+    av->prev = *tail;
+    *tail = av;
+
+    return av;
+error:
+    meta_freeavlist(*head); *head = *tail = 0; 
+    return 0;
+}
+
+
 /* Decodes a data block, forming a flat list of AV items. */
 
 META_AV *meta_decode(META *m, META_SPC *s, META_ORD defvnd, 
@@ -294,20 +336,9 @@ META_AV *meta_decode(META *m, META_SPC *s, META_ORD defvnd,
 
 			/* Add the data, unless we shouldn't decode it */
 			if (!i->nodec) {
-
-				av = (META_AV *)malloc(sizeof(META_AV));
-				if (!av) { meta_freeavlist(avlist); return 0; }
-
-				memset(av, 0, sizeof(META_AV));
-				av->i = i;
-				if (MT_ISORD(i->val_type))
-					av->l = getord(blk+i->val_ofs, vallen);
-				else av->l = vallen, av->p = blk+i->val_ofs;
-
-				if (!avlist) avlist = av;
-				if (tail) tail->next = av;
-				av->prev = tail;
-				tail = av;
+			    if (!addavfromraw(&avlist, &tail, i,
+					      blk + i->val_ofs, vallen,
+					      copystrings)) return 0;
 			}
 
 			/* Recurse into a subspace, if any */
@@ -355,32 +386,9 @@ META_AV *meta_decode(META *m, META_SPC *s, META_ORD defvnd,
 
 		/* Add the data, unless we shouldn't decode it */
 		if (!i->nodec) {
-
-			av = (META_AV *)malloc(sizeof(META_AV));
-			if (!av) { meta_freeavlist(avlist); return 0; }
-
-			memset(av, 0, sizeof(META_AV));
-			av->i = i;
-			if (MT_ISORD(i->val_type))
-				av->l = getord(p + i->val_ofs, vallen);
-			else if (!copystrings) {
-				av->l = vallen; 
-				av->p = p + i->val_ofs;
-			} else {
-				/* Create a copy for all strings on the heap
-				 * instead of having them point in the source 
-				 * buffer. Used by langvm when DICTDEC uses
-				 * a temporary string as the source. */
-				av->l = vallen; 
-				av->p = (char *)malloc(av->l);
-				if (!av->p) { free(av); meta_freeavlist(avlist); return 0; }
-				memcpy(av->p, p + i->val_ofs, av->l); 
-				av->flags = AV_FREE_P; 
-			}
-			if (!avlist) avlist = av;
-			if (tail) tail->next = av;
-			av->prev = tail;
-			tail = av;
+		    if (!addavfromraw(&avlist, &tail, i,
+				      p + i->val_ofs, vallen,
+				      copystrings)) return 0;
 		}
 
 		/* Recurse into a subspace, if any */
@@ -757,12 +765,13 @@ void meta_buildtree(META_AV *src, META_AV **dstlst, META_SPC *groundspc)
 			 * encapsulating attributes (recursively). There's no
 			 * encapsulation route possible, so skip this item. */
 
-			msg(F_SEND, L_DEBUG, "meta_buildtree: skipping %s::%s "
-					     "(%ld:%ld:%ld): no encapsulation "
-					     "route to space %s (%ld)\n",
-			    srcav->i->spc->name, srcav->i->name,
-			    srcav->i->spc->nr, srcav->i->vnd, srcav->i->nr,
-			    groundspc->name, groundspc->nr);
+			D1(msg(F_SEND, L_DEBUG, "meta_buildtree: "
+			                        "skipping %s::%s (%ld:%ld:%ld):"
+						" no encapsulation route to "
+						"space %s (%ld)\n",
+			       srcav->i->spc->name, srcav->i->name,
+			       srcav->i->spc->nr, srcav->i->vnd, srcav->i->nr,
+			       groundspc->name, groundspc->nr));
 
 			free(srcav);
 			srcav = getcopy(&src);
@@ -994,9 +1003,7 @@ int meta_binmsgtoav(META *m, U_INT32_T *buf, ssize_t len,
 		vndnr = netint32(*i); i++;
 		atrnr = netint32(*i); i++;
 		atrlen = netint32(*i); i++;
-		msg(F_PROC, L_DEBUG, "- got spcnr %ld, vndnr %ld, atrnr %ld, "
-				     "atrlen %ld\n",
-		    (long)spcnr, (long)vndnr, (long)atrnr, (long)atrlen);
+		D1(msg(F_PROC, L_DEBUG, "- got spcnr %ld, vndnr %ld, atrnr %ld, atrlen %ld\n", (long)spcnr, (long)vndnr, (long)atrnr, (long)atrlen));
 
 		spc = meta_getspcbynr(m, spcnr);
 		if (!spc) { msg(F_PROC, L_NOTICE, "meta_binmsgtoav: Ignoring "
@@ -1043,10 +1050,8 @@ int meta_binmsgtoav(META *m, U_INT32_T *buf, ssize_t len,
 		}
 
 		/* Add A/V item to list */
-		if (msg_thresh[F_PROC] >= L_DEBUG) meta_printav(m, av, 0);
 		meta_addav(head, tail, 0, 0, av);
 	}
-
 	return 0;
 }
 
@@ -1267,6 +1272,16 @@ ssize_t meta_avtomsg(META *m, META_AV *avlst,
 			*o++ = '"';
 		}
 
+#if 0			    /* debug memory mgmt; only printav sets ONESHOT */
+		if (flags & AVMSG_ONESHOT) {
+		    buflen -= 13; if (buflen < 0) break;
+		    *o++ = ' '; *o++ = '(';
+		    *o++ = '0'; *o++ = 'x';
+		    hex(o, &av->p, 4); o += 8;
+		    *o++ = ')';
+		}
+#endif
+
 		/* Write LF */
 		buflen--; if (buflen < 0) break; 
 		*o++ = '\n';
