@@ -369,12 +369,18 @@ int vm_run(META *m, VM *vm)
 			meta_freeavdata(&R); R.p = 0; R.l = 0;
 			if (!R.i) break;
 
+			/* Show source list */
+			if (msg_thresh[F_SEND] >= L_DEBUG) {
+				msg(F_SEND, L_DEBUG, "OP_DICTENC: encoding %s list:\n", R.ref.list == AVR_LIST_REP ? "REP" : "REQ");
+				meta_printavlist(m, vm->head[R.ref.list], 0);
+			}
+
 			/* Build a tree from the encapsulating space */
 			avlst = 0;
 			meta_buildtree(vm->head[R.ref.list], &avlst, R.i->spc);
 			if (!avlst) break;
 			if (msg_thresh[F_SEND] >= L_DEBUG) {
-				msg(F_SEND, L_DEBUG, "OP_DICTENC: tree for %s list from %s:\n", R.ref.list == AVR_LIST_REP ? "REP" : "REQ", R.i->spc->name);
+				msg(F_SEND, L_DEBUG, "OP_DICTENC: tree on ground space %s:\n", R.i->spc->name);
 				meta_printavlist(m, avlst, 0);
 			}
 			/* Now encode it */
@@ -382,6 +388,7 @@ int vm_run(META *m, VM *vm)
 			R.l = meta_encode(R.i->spc, s, C_MAX_PKTSIZE, avlst, 0);
 			if (R.l < 0) { R.l = 0; free(s); break; }
 			R.p = s; R.i = 0; R.flags |= AV_FREE_P;
+			if (msg_thresh[F_SEND] >= L_DEBUG) hexdump(s, R.l);
 			break;
 
 		  /*
@@ -559,6 +566,10 @@ int vm_run(META *m, VM *vm)
 				meta_freeavdata(&L); L.l = 0; L.p = 0; L.i = 0; 
 				break; 
 			}
+			if (msg_thresh[F_RECV] >= L_NOTICE) {
+				msg(F_RECV, L_NOTICE, "OP_DICTDEC: %spairs added to %s of %s list:\n", n ? "copied " : "", R.ref.inst == AVR_INST_FIRST ? "top" : "bottom", R.ref.list == AVR_LIST_REP ? "REP" : "REQ");
+				meta_printavlist(m, avlst, 0);
+			}
 			/* Add avlst/av (head/tail) to proper place */
 			if (R.ref.inst == AVR_INST_FIRST) {	/* head */
 				av->next = vm->head[R.ref.list];
@@ -573,6 +584,9 @@ int vm_run(META *m, VM *vm)
 			if (avlst->prev) avlst->prev->next = avlst;
 			if (!vm->head[R.ref.list])
 				vm->head[R.ref.list] = avlst;
+
+			/* Return success */
+			meta_freeavdata(&L); L.l = 1; L.p = 0; L.i = 0;
 			break;
 
 		  /*
@@ -970,18 +984,22 @@ int vm_run(META *m, VM *vm)
 			break;
 
 		  case OP_JOINSTR: if (sp < 1) { ret = VM_UNDERFLOW; break; } 
+			/* We only operate on attribute references, 
+			 * not anonymous values */
 			if (!R.i) break;
 
-			/* First free right operand's data if any - this occurs
-			 * if you do tricks like all (str . "a"); we only care
-			 * about the reference, not the stack item's data. */
+			/* First free right operand's data if any - this must
+			 * be done if you do tricks like all (str . "a"); we
+			 * only care about the underlying item, not the stack
+			 * item's data. */
 			meta_freeavdata(&R); R.p = 0; R.l = 0;
 
 			/* First get length of all pairs. This operator breaks
 			 * the exception to the 'no late binding' rule (because
 			 * of the type switch here), but we can't use the 
-			 * compiler's autoconversion here because we haven't
-			 * got proper lists implemented yet. */
+			 * compiler's autoconversion here because the auto-
+			 * conversion opcode would only do one item. FIXME:
+			 * we should probably add a separate JOINORD instead. */
 			n = 0;
 			if (MT_ISORD(R.i->val_type)) {
 				for(av = vm->head[R.ref.list]; 
